#!/usr/bin/env python3

import numpy as np
from scipy.misc import imread, imshow, imsave


def image2pixels(image):
    """Create a list of (R, G, B, X, Y) tuples from a given 3d image array."""
    result = []
    y = image.shape[0]
    x = image.shape[1]
    for i in range (0, y):
        for j in range (0, x):
            a = image[i][j]
            meow = (a[0],a[1],a[2],j,i)
            result.append(meow)
    return result


def pixels2image(pixels):
    """Create a 3d image array from a list of (R, G, B, X, Y) pixels."""
    xmax = 0
    ymax = 0
    for i in pixels:
        if i[3] > xmax:
            xmax = i[3]
        if i[4] > ymax:
            ymax = i[4]
    result = np.zeros((ymax+1,xmax+1,3), dtype='uint8')
    for i in pixels:
        for k in range(0,3):
            result[i[4]][i[3]][k] = i[k]
    return result


def bounding_box(pixels):
    """Return a tuple ((Rmin, Rmax), (Gmin, Gmax), (Bmin, Bmax)) with the
    respective minimum and maximum values of each color."""
    rmin = 255
    rmax = 0
    gmin = 255
    gmax = 0
    bmin = 255
    bmax = 0
    for i in pixels:
        if i[0] < rmin:
                rmin = i[0]
        if i[0] > rmax:
                rmax = i[0]
        if i[1] < gmin:
                gmin = i[1]
        if i[1] > gmax:
                gmax = i[1]
        if i[2] < bmin:
                bmin = i[2]
        if i[2] > bmax:
                bmax = i[2]
    return ((rmin,rmax),(gmin,gmax),(bmin,bmax))


def color_average(pixels):
    """Return list of tuples (Ravg, Gavg, Bavg, X, Y) with averaged color values."""
    listlength = len(pixels)
    if (listlength == 0):
        return []
    rges = 0
    gges = 0
    bges = 0
    for i in pixels:
        rges += i[0]
        gges += i[1]
        bges += i[2]
    ravg = round(rges/listlength)
    gavg = round(gges/listlength)
    bavg = round(bges/listlength)
    a = []
    for i in pixels:
        x = (ravg, gavg, bavg, i[3], i[4])
        a.append(x)
    return a

def cut_dimension(bbox):
    """Determine the longest side of the bounding box, i.e. return 0, 1 or 2. In case of a draw, prefer
    the lower dimensions."""
    rlength = bbox[0][1] - bbox[0][0]
    glength = bbox[1][1] - bbox[1][0]
    blength = bbox[2][1] - bbox[2][0]
    if rlength >= glength and rlength >= blength:
        return 0
    if glength > rlength and glength >= blength:
        return 1
    return 2


def recursive_median_cut(pixels, N, bbox=False):
    if N == 0 or len(pixels) < 2:
        return color_average(pixels)
    count = len(pixels)
    half = int(count/2)
    box = bounding_box(pixels)
    cut = cut_dimension(box)
    pixels.sort(key=lambda x: x[cut])
    # split it...
    a = recursive_median_cut(pixels[0:half], N-1)
    b = recursive_median_cut(pixels[half:count], N-1)
    return a+b

def median_cut(image, ncuts=8):
    """Perform the median cut algorithm on a given image. Image may be a
filename, or a three dimensional array of RGB values."""
    pixels = image2pixels(image)
    pixels = recursive_median_cut(pixels, ncuts)
    return pixels2image(pixels)


def image_difference(imageA, imageB):
    """Compute the difference of two images in the maximum-norm."""
    if imageA.shape != imageB.shape:
        raise ValueError('Dimensions of both images must be equal!')
    diff = np.array(imageA,dtype='int16') - np.array(imageB,dtype='int16')
    return np.amax(abs(diff))


def main():
    shibuya = imread('shibuya.png', mode='RGB')
    lena    = imread('lena.png', mode='RGB')
    # median cut quantisation with 2^N colors
    N = 4
    shibuya2 = median_cut(shibuya, N)
    lena2 = median_cut(lena, N)
    print("shibuya 2^{} colors: {}".format(N, image_difference(shibuya, shibuya2)))
    print("lena    2^{} colors: {}".format(N, image_difference(lena, lena2)))
    imshow(lena2)
    imshow(shibuya2)


if __name__ == "__main__": main()
