#!/usr/bin/env python3

import numpy as np
import matplotlib.cm as cm
import matplotlib.pyplot as plt
import matplotlib.animation as animation

weights = np.array([1./36, 1./9, 1./36,
                    1./9,  4./9, 1./9,
                    1./36, 1./9, 1./36])

directions = np.array([(-1,-1), ( 0,-1), (1,-1),
                       (-1, 0), ( 0, 0), (1, 0),
                       (-1, 1), ( 0, 1), (1, 1)])

inverse_dir = np.array([8, 7, 6, 5, 4, 3, 2, 1, 0])

def density(f):
    return np.sum(f, axis = 0)

def velocity(f):
    ux = np.zeros((f.shape[1],f.shape[2])) # w x h array
    uy = np.zeros((f.shape[1],f.shape[2])) # w x h array
    for i in range(0,9):
        for j in range(0,f.shape[1]):
            for k in range(0,f.shape[2]):
                ux[j][k] += directions[i][0] * f[i][j][k]
                uy[j][k] += directions[i][1] * f[i][j][k]
    return (ux/density(f), uy/density(f))

def f_eq(f):
    a = velocity(f)
    ux = a[0]
    uy = a[1]
    res = []
    ro = density(f)
    for i in range (0,9):
        x = weights[i] * ro * (1+3* (directions[i][0]*ux + directions[i][1]*uy) 
                + 9./2 * (directions[i][0]*ux + directions[i][1]*uy)**2
                - 3./2 * (ux*ux + uy*uy))
        res.append(x)
    return np.asarray(res)


def collide(f, omega):
    return f - omega*(f - f_eq(f))


def stream(f):
    f_old = f.copy()
    f_new = f.copy()
    for i in range(1,f.shape[1]-1):
        for j in range(1,f.shape[2]-1):
            # propagate values
#            f_new[0][i-1][j-1] = f_old[0][i][j]
#            f_new[1][i-1][j]   = f_old[1][i][j]
#            f_new[2][i-1][j+1] = f_old[2][i][j]
#
#            f_new[3][i][j-1] = f_old[3][i][j]
#            f_new[5][i][j+1] = f_old[5][i][j]
#
#            f_new[6][i+1][j-1] = f_old[6][i][j]
#            f_new[7][i+1][j]   = f_old[7][i][j]
#            f_new[8][i+1][j+1] = f_old[8][i][j]

            #substitute
            f_new[0][i][j] = f_old[0][i+1][j+1]
            f_new[1][i][j] = f_old[1][i][j+1]
            f_new[2][i][j] = f_old[2][i-1][j+1]
            f_new[3][i][j] = f_old[3][i+1][j]
            f_new[5][i][j] = f_old[5][i-1][j]
            f_new[6][i][j] = f_old[6][i+1][j-1]
            f_new[7][i][j] = f_old[7][i][j-1]
            f_new[8][i][j] = f_old[8][i-1][j-1]
    return f_new


def noslip(f, masklist):
    pass # TODO
    return f


flow = np.array([1./36, 1./9, 1.6/36,
                 1./9,  4./9, 1.6/9,
                 1./36, 1./9, 1.6/36])

def lbm(W, H, timesteps=1000, omega=1.85):
    fig, ax = plt.subplots()
    f = weights.reshape((9,1,1)) * np.ones((9, W, H))
    f[:, 0, :] = flow.reshape(9,1)
    # create the list of obstacles
    mask = np.zeros((W,H),dtype=bool)
    mask[int(0.2*W):int(0.22*W), int(0.4*H):int(0.6*H)] = 1
    mask[1:-1, 1] = 1
    mask[1:-1,-2] = 1
    masklist = np.argwhere(mask)

    (ux, uy) = velocity(f)
    mat = ax.matshow((np.sqrt(ux*ux + uy*uy)).transpose(), cmap=cm.GnBu)
    def update_velocity_field():
        nonlocal f
        for rep in range(10):
            # inflow
            f[:, 0, :] = flow.reshape(9,1)
            # copying outflow
            f[:, -2, :] = f[:, -1, :]
            f = stream(noslip(collide(f, omega), masklist))
        (ux, uy) = velocity(f)
        mat.set_data((np.sqrt(ux*ux + uy*uy)).transpose())
    ani = animation.FuncAnimation(fig, lambda _: update_velocity_field(),
                                  frames=10000,
                                  interval=10,
                                  blit=False)

    plt.show()

def main():
    lbm(300, 60)


if __name__ == "__main__": main()
